/*
     This file is part of GNUnet

     GNUnet is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published
     by the Free Software Foundation; either version 2, or (at your
     option) any later version.

     GNUnet is distributed in the hope that it will be useful, but
     WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with GNUnet; see the file COPYING.  If not, write to the
     Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
*/

/**
 * @file src/applications/afs/gtkui/insert.c
 * @author Igor Wronsky
 * @author Christian Grothoff (refactoring, added bugs)
 *
 * This handles file insertions in the GUI
 **/

#include "config.h"

#include <errno.h>
#include <gtk/gtk.h>
#include <string.h>

#include "util/checksum.h"
#include "util/configuration.h"
#include "util/returnvalues.h"
#include "util/xmalloc.h"
#include "applications/afs/encoding/insertutil.h"
#include "applications/afs/encoding/keywords.h"
#include "applications/afs/gtkui/helper.h"
#include "applications/afs/gtkui/insert.h"
#include "applications/afs/gtkui/main.h"
#include "util/port.h"
#include "util/storage.h"
#include "util/logging.h"

/**
 * This structure works as a placeholder for almost
 * anything that is needed during the insertion process. Blah.
 **/
typedef struct {
  char * fileName;
  char * fileNameRoot;
  char * description;
  char * mimetype;
  char ** keywords;
  int num_keywords;
  int indexContent;
  GtkWidget * progressBar;
  GtkWidget * progressBarWindow;
} InsertModel;

/**
 * This structure works as a placeholder for almost
 * anything that is needed during the insertion process. Blah.
 **/
typedef struct {
  char * fileName;
  char * mimetype; /* FIXME: add GtkWidget for editing instead! */
  GtkWidget * editAttributesWindow;
  GtkWidget * fileNameLine;
  GtkWidget * descriptionLine;
  GtkWidget * indexButton;
  GtkWidget * keywordLine;
  GtkWidget * keywordList;
} EditWindowModel;

/**
 * Callback function to update the insert progressbar.
 *
 * @param stats Statistics related to insert progression
 * @param ilm Data related to the insertion
 **/
static void insertModelCallback(ProgressStats * stats,
				InsertModel * ilm) {
  gdk_threads_enter();
  gtk_progress_set_value(GTK_PROGRESS(ilm->progressBar),
			 stats->progress);
  gdk_threads_leave();
}


/**
 * A function to be run by the insert thread. Does the
 * actual insertion.
 *
 * @param ilm Collected data related to the insertion
 */
static void insertFile_(InsertModel * ilm) {
  int res;
  GNUNET_TCP_SOCKET * sock;
  Block * top;
  int i;
  
  SEMAPHORE_DOWN(refuseToDie);
  sock = getClientSocket();
  top = insertFile(sock,
		   ilm->fileName,
		   (ProgressModel)&insertModelCallback,
		   ilm);
  if (top != NULL) {
    res = insertRoot(sock,
		     top, 
		     ilm->description,
		     ilm->fileNameRoot,
		     ilm->mimetype,
		     ilm->num_keywords,
		     ilm->keywords);
  } else {
    res = SYSERR;
  }
  gdk_threads_enter();
  gtk_widget_destroy(ilm->progressBarWindow);
  if (res == OK) {
    HexName hex;
    HexName hex2;
    
    hash2hex(&top->chk.key,&hex);
    hash2hex(&top->chk.query, &hex2);

    guiMessage("Successfully processed %s\n%s\n%s\n%u %u\n", 
	       ilm->fileName,
	       (char*)&hex,
	       (char*)&hex2,
	       crc32N(top->data, top->len),
	       top->filesize);
    LOG(LOG_DEBUG,
	"DEBUG: Successfully processed %s -- %s %s %u %u\n", 
	ilm->fileName,
	(char*)&hex,
	(char*)&hex2,
	crc32N(top->data, top->len),
	top->filesize);
  } else {
    guiMessage("Insertion of %s FAILED!\n", 
	       ilm->fileName);
  }
  gdk_threads_leave();
  if(top != NULL)
    top->done(top, NULL);
  releaseClientSocket(sock);

  /* insert complete */
  SEMAPHORE_UP(refuseToDie);
  for (i=0;i<ilm->num_keywords;i++)
    FREE(ilm->keywords[i]);
  FREENONNULL(ilm->keywords);
  FREE(ilm->fileName);
  FREE(ilm->mimetype);
  FREE(ilm->description);
  FREE(ilm->fileNameRoot);
  FREE(ilm);
}

/**
 * Callback for handling "delete_event": keep the window OPEN.
 **/
gint refuseDeleteEvent(GtkWidget * widget,
		       GdkEvent * event,
		       gpointer data) {
  LOG(LOG_DEBUG, 
      "DEBUG: refuseDeleteEvent\n");
  return TRUE;
}

/**
 * Collects the results of editAttributes, creates an insertion 
 * progressbar and launches the insertion thread.
 *
 * @param ilm Data related to the insertion
 **/
static void startInsert(GtkWidget * dummy, 
			EditWindowModel * ewm) {
  GtkWidget * window;
  GtkWidget * box;
  GtkObject * adjustment;
  InsertModel * ilm;
  gchar * txt;
  int i;
  pthread_t insertThread;
  int fileLength;

  ilm = MALLOC(sizeof(InsertModel));
  ilm->fileName = STRDUP(ewm->fileName);

  /* get content indexing style */
  if(gtk_toggle_button_get_active((GtkToggleButton *)ewm->indexButton) == TRUE)
    ilm->indexContent = YES;
  else
    ilm->indexContent = NO;
  
  /* get the published filename */
  txt = gtk_entry_get_text(GTK_ENTRY(ewm->fileNameLine));
  if (txt == NULL)
    ilm->fileNameRoot = STRDUP("none specified");
  else
    ilm->fileNameRoot = STRDUP(txt);
  
  /* get the new description, if any */
  txt = gtk_entry_get_text(GTK_ENTRY(ewm->descriptionLine));
  if (txt == NULL)
    ilm->description = STRDUP("no description specified");
  else
    ilm->description = STRDUP(txt);

  ilm->mimetype = STRDUP(ewm->mimetype);

  /* get new list of keywords */
  ilm->num_keywords = GTK_CLIST(ewm->keywordList)->rows;

  if (ilm->num_keywords > 0) {
    ilm->keywords = (char**) MALLOC(ilm->num_keywords * sizeof(char*));
    for(i=0;i<ilm->num_keywords;i++) {     
      gtk_clist_get_text(GTK_CLIST(ewm->keywordList),
			 i,
			 0,
			 &txt);
      ilm->keywords[i] = STRDUP(txt);
    } 
  } else
    ilm->keywords = NULL;

  /* create a new window for a progressbar */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  ilm->progressBarWindow = window;
  gtk_window_set_title(GTK_WINDOW(window), 
		       gtk_entry_get_text(GTK_ENTRY(ewm->fileNameLine)));
  box = gtk_hbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(window), 
		    box);
  gtk_signal_connect(GTK_OBJECT(window),
		     "delete_event",
		     GTK_SIGNAL_FUNC(refuseDeleteEvent),
		     NULL);
  gtk_container_set_border_width(GTK_CONTAINER(window), 
				 10);

 
  /* create the actual progressbar */
  fileLength = getFileSize(ilm->fileName);
  ilm->progressBar = gtk_progress_bar_new();
  gtk_progress_set_show_text(GTK_PROGRESS(ilm->progressBar),
			     1);
  gtk_progress_set_format_string(GTK_PROGRESS(ilm->progressBar),
				 (gchar*)"%v bytes inserted");
  adjustment = gtk_adjustment_new(0,
				  0,
				  fileLength,
				  1,
				  0,
				  0);
  gtk_progress_set_adjustment(GTK_PROGRESS(ilm->progressBar),
			      GTK_ADJUSTMENT(adjustment));
  gtk_box_pack_start(GTK_BOX(box),
		     ilm->progressBar, 
		     TRUE, 
		     TRUE, 
		     0);
  gtk_progress_bar_set_orientation(GTK_PROGRESS_BAR(ilm->progressBar),
				   GTK_PROGRESS_LEFT_TO_RIGHT);
  gtk_widget_show(ilm->progressBar);
  gtk_widget_show(box);
  gtk_widget_show(window);
 
  /* start the insert thread */
  if (0 != PTHREAD_CREATE(&insertThread,
			  (void * (*)(void *)) insertFile_,
			  ilm,
			  16 * 1024))
    errexit("FATAL: could not create insert thread (%s)!\n",
	    strerror(errno));
  pthread_detach(insertThread);

  /* destroy the "editAttributes" window */
  gtk_widget_destroy(ewm->editAttributesWindow);
}

/**
 * The keyword add button was clicked. Add whatever 
 * is in the keyword box to the list of keywords.
 *
 * @param ilm the insert model
 **/
static void button_add_clicked(GtkWidget * w, 
			       EditWindowModel * ewm) {
  gchar * key;
  gchar * newKeyword;
  int i;

  key = gtk_entry_get_text(GTK_ENTRY(ewm->keywordLine));
  if (key == NULL) {
    /* message to enter a string? */
    return;
  }    

  newKeyword = STRDUP(key);
  key = newKeyword;

  /* remove trailing & heading spaces */
  i = strlen(key)-1;
  while ( (newKeyword[i] == ' ') && 
	  (i >= 0) ) {
    newKeyword[i--] = '\0';
  }
  while (*newKeyword == ' ')
    newKeyword++;

  if ( *newKeyword == '\0' ) {
    /* message to enter more than spaces? */    
  } else {
    gtk_clist_append(GTK_CLIST(ewm->keywordList),
		     &newKeyword);
  } 
  FREE(key);
  gtk_entry_set_text(GTK_ENTRY(ewm->keywordLine),
		     "");
}

/**
 * The keyword delete button was clicked. Delete the 
 * currently selected keyword.
 *
 * @param ilm the insert model
 **/
static void button_del_clicked(GtkWidget * w, 
			       EditWindowModel * ewm) {
  GList * tmp;

  tmp = GTK_CLIST(ewm->keywordList)->selection;
  if (NULL == tmp) {
    /* message that keyword must be selected to delete one? */
    return;
  }  
  gtk_clist_remove(GTK_CLIST(ewm->keywordList),
		   (int)tmp->data);
}

/**
 * Exit the application (called when the main window
 * is closed or the user selects File-Quit).
 **/
static void destroyEditWindow(GtkWidget * widget,
			      EditWindowModel * ewm) {
  FREE(ewm->fileName);
  FREE(ewm->mimetype);
  FREE(ewm);
}

/**
 * Show user a window to edit information related to this file.
 * After user is done, call startInsert.
 *
 * @param lots of data related to the insertion
 **/
static void editAttributes(char * filename,
			   char * fileNameRoot,
			   char * description,
			   char * mimetype,
			   char ** keywords,
			   int num_keywords) {
  EditWindowModel * ewm;
  GtkWidget * window;
  GtkWidget * vbox, * hbox;
  GtkWidget * clist;
  GtkWidget * scrolled_window;
  GtkWidget * label;
  GtkWidget * separator; 
  GtkWidget * button_add;
  GtkWidget * button_delete;
  GtkWidget * button_ok;
  GtkWidget * button_cancel;
  GtkWidget * keyword_line;
  GSList * group;
  GtkWidget * indexbutton1;
  GtkWidget * indexbutton2;
  gchar * titles[1] = { "Keyword(s) used" };
  int i;

  ewm = MALLOC(sizeof(EditWindowModel));
  ewm->fileName = STRDUP(filename);
  if (mimetype == NULL)
    ewm->mimetype = STRDUP("unknown");
  else
    ewm->mimetype = STRDUP(mimetype);
  /* create new window for editing */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  ewm->editAttributesWindow = window;
  gtk_widget_set_usize(GTK_WIDGET(window),
		       400,
		       400);
  gtk_window_set_title(GTK_WINDOW(window), 
		       "Edit attributes");

  /* add container for window elements */
  vbox = gtk_vbox_new(FALSE, 0);
  gtk_container_add(GTK_CONTAINER(window),
		    vbox);
  gtk_widget_show(vbox);

  /* when user clicks on close box, always "destroy" */
  gtk_signal_connect(GTK_OBJECT(window),
		     "delete_event",
		     GTK_SIGNAL_FUNC(deleteEvent),
		     ewm);
  /* whenever edit window gets destroyed, 
     free *ALL* ewm data */
  gtk_signal_connect(GTK_OBJECT(window),
		     "destroy",
		     GTK_SIGNAL_FUNC(destroyEditWindow),
		     ewm);

  gtk_container_set_border_width(GTK_CONTAINER(window), 
				 10);

  /* Create a line to change the published filename */
  label = gtk_label_new("Published filename:");
  gtk_box_pack_start(GTK_BOX(vbox),
		     label, 
		     FALSE, 
		     FALSE, 
		     0);
  gtk_widget_show(label); 
  ewm->fileNameLine = gtk_entry_new();
  gtk_box_pack_start(GTK_BOX(vbox),
		     ewm->fileNameLine,
		     TRUE,
		     TRUE,
		     0);
  gtk_entry_set_text(GTK_ENTRY(ewm->fileNameLine), 
		     fileNameRoot);
  gtk_widget_show(ewm->fileNameLine);
  
  /* Create a line to change the description */
  label = gtk_label_new("Description:");
  gtk_box_pack_start(GTK_BOX(vbox),
		     label, 
		     FALSE, 
		     FALSE, 
		     0);
  gtk_widget_show(label);  
  ewm->descriptionLine = gtk_entry_new();
  gtk_box_pack_start(GTK_BOX(vbox),
		     ewm->descriptionLine, 
		     TRUE, 
		     TRUE,
		     0);
  gtk_entry_set_text(GTK_ENTRY(ewm->descriptionLine), 
		     description);
  gtk_widget_show(ewm->descriptionLine);
  
  /* add buttons to select the insertion method */
  label = gtk_label_new("Insertion method:");
  gtk_box_pack_start(GTK_BOX(vbox),
		     label,
		     FALSE, 
		     FALSE,
		     0);
  gtk_widget_show(label);  
  hbox = gtk_hbox_new(FALSE,0);
  gtk_box_pack_start(GTK_BOX(vbox), 
		     hbox, 
		     TRUE,
		     TRUE,
		     0);
  gtk_widget_show(hbox);
  indexbutton1 = gtk_radio_button_new_with_label(NULL,
						 "Index only");
  ewm->indexButton = indexbutton1;
  gtk_box_pack_start(GTK_BOX (hbox),
		     indexbutton1,
		     TRUE,
		     TRUE,
		     0);
  gtk_widget_show(indexbutton1);  
  group = gtk_radio_button_group(GTK_RADIO_BUTTON(indexbutton1));
  indexbutton2 = gtk_radio_button_new_with_label(group, 
						 "Full insertion");
  gtk_box_pack_start(GTK_BOX(hbox), 
		     indexbutton2, 
		     TRUE,
		     TRUE,
		     0);
  gtk_widget_show(indexbutton2);
  if (testConfigurationString("GNUNET-INSERT",
  			      "INDEX-CONTENT",
			      "YES") == YES)
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(indexbutton1), 
				 TRUE);
  else
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(indexbutton2), 
				 TRUE);

  /* add a list of keywords */
  scrolled_window = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window),
				 GTK_POLICY_AUTOMATIC, 
				 GTK_POLICY_ALWAYS);
  gtk_box_pack_start(GTK_BOX(vbox), 
		     scrolled_window, 
		     TRUE, 
		     TRUE, 
		     0);
  gtk_widget_show(scrolled_window);  
  clist = gtk_clist_new_with_titles(1, titles); 
  ewm->keywordList = clist;
  gtk_container_add(GTK_CONTAINER(scrolled_window), 
		    clist);

  /* add the pre-extracted keywords to the list */
  gtk_clist_freeze(GTK_CLIST(clist));
  for(i=0;i<num_keywords;i++) {
    gtk_clist_append(GTK_CLIST(clist), 
		     &keywords[i]);
  }
  gtk_clist_thaw(GTK_CLIST(clist));
  gtk_widget_show(clist);

  /* add a line to input new keywords */
  keyword_line = gtk_entry_new();
  ewm->keywordLine = keyword_line;
  gtk_box_pack_start(GTK_BOX(vbox),
		     keyword_line, 
		     FALSE,
		     FALSE,
		     0);
  gtk_entry_set_text(GTK_ENTRY(keyword_line), 
		     "");
  gtk_signal_connect(GTK_OBJECT(keyword_line),
		     "activate",
                     GTK_SIGNAL_FUNC(button_add_clicked),
                     ewm);
  gtk_widget_show(keyword_line);

  /* add the buttons to add and delete keywords */
  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox),
		     hbox,
		     FALSE,
		     TRUE,
		     0);
  gtk_widget_show(hbox);
  button_add = gtk_button_new_with_label("Add keyword");
  button_delete = gtk_button_new_with_label("Delete keyword");
  gtk_box_pack_start(GTK_BOX(hbox), 
		     button_add, 
		     TRUE, 
		     TRUE, 
		     0);
  gtk_box_pack_start(GTK_BOX(hbox), 
		     button_delete, 
		     TRUE, 
		     TRUE, 
		     0);
  gtk_signal_connect(GTK_OBJECT(button_add),
		     "clicked",
		     GTK_SIGNAL_FUNC(button_add_clicked),
		     ewm);
  gtk_signal_connect(GTK_OBJECT(button_delete), 
		     "clicked",
		     GTK_SIGNAL_FUNC(button_del_clicked),
		     ewm);
  gtk_widget_show(button_add);
  gtk_widget_show(button_delete);

  /* add the insertion ok/cancel buttons */
  separator = gtk_hseparator_new();
  gtk_box_pack_start(GTK_BOX(vbox),
		     separator,
		     TRUE, 
		     TRUE,
		     0);
  gtk_widget_show(separator);

  hbox = gtk_hbox_new(FALSE, 0);
  gtk_box_pack_start(GTK_BOX(vbox),
		     hbox, 
		     FALSE, 
		     TRUE, 
		     0);
  gtk_widget_show(hbox);
  button_ok = gtk_button_new_with_label("Ok");
  button_cancel = gtk_button_new_with_label("Cancel");
  gtk_box_pack_start(GTK_BOX(hbox),
		     button_ok,
		     TRUE,
		     TRUE,
		     0);
  gtk_box_pack_start(GTK_BOX(hbox), 
		     button_cancel, 
		     TRUE,
		     TRUE, 
		     0);
  gtk_signal_connect(GTK_OBJECT(button_ok), 
		     "clicked",
		     GTK_SIGNAL_FUNC(startInsert),
		     ewm);
  gtk_signal_connect(GTK_OBJECT(button_cancel),
		     "clicked",
		     GTK_SIGNAL_FUNC(destroyWidget),
		     window);
  gtk_widget_show(button_ok);
  gtk_widget_show(button_cancel);

  /* all clear, show the window */
  gtk_widget_show(window);
}

/**
 * Callback for the file selection window. Launches the attribute
 * edit routine for the selected file, after keywords have been
 * extracted.
 *
 * @param window the file selection window
 */
static gint file_selected(GtkWidget * okButton, 
			  GtkWidget * window) {
  gchar * filename;
  int i;
  char * fileNameRoot;
  char * description;
  char * mimetype;
  char ** keywords;
  int num_keywords; 

  filename 
    = gtk_file_selection_get_filename(GTK_FILE_SELECTION(window));
  if ( (filename == NULL) ||
       (0 == assertIsFile(filename)) ) {
    guiMessage("Please select a file!\n");
    gtk_widget_destroy(window);
    return FALSE;
  }

  filename = STRDUP(filename);

  /* if filename is '/home/user/foo', use 'foo' as the filenameRoot */
  fileNameRoot = NULL;
  for (i=strlen(filename)-1;i>=0;i--) {
    if (filename[i] == '/') {
      fileNameRoot = STRDUP(&filename[i+1]);
      break;
    }
  }  
  if (i == -1)
    errexit("FATAL: ASSERTION failed: filename does not start with a '/'");

  /* try to extract keywords */ 
  description = NULL;
  mimetype = NULL; 
  num_keywords = 0;
  keywords = NULL;
  extractKeywords(filename,
		  &description,
		  &mimetype,
		  &keywords,
		  &num_keywords);

  if (description == NULL)
    description = STRDUP("No description supplied");
  if (mimetype == NULL)
    mimetype = STRDUP("unknown");
  
  /* destroy the open-file window */
  gtk_widget_destroy(window);

  /* allow the user to edit the insertion related info */
  editAttributes(filename,
		 fileNameRoot,
		 description,
		 mimetype,
		 keywords,
		 num_keywords);

  for (i=0;i<num_keywords;i++)
    FREE(keywords[i]);
  FREENONNULL(keywords);
  FREE(mimetype);
  FREE(description);
  FREE(filename);
  FREE(fileNameRoot);
  return FALSE;
}

/**
 * Close the open-file window.
 **/
static gint destroyOpenFile(GtkWidget * widget,
			    GtkWidget * window) {
  LOG(LOG_DEBUG, 
      "DEBUG: destroying open-file window (%x)\n", 
      window);
  return TRUE;
}

/**
 * Pops up a file selector for the user.
 *
 * Explanation: In insertion, functions will be called in the
 * following order,
 *
 * openSelectFile[OK click->]file_selected->editAttributes[OK click]
 * ->startInsert->newthread.insertFile_
 **/
void openSelectFile(void) {
  GtkWidget * window;

  window = gtk_file_selection_new("Choose file to be inserted");
  gtk_signal_connect(GTK_OBJECT(window), 
		     "destroy",
		     GTK_SIGNAL_FUNC(destroyOpenFile),
		     window);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(window)->ok_button),
		     "clicked", 
		     GTK_SIGNAL_FUNC(file_selected),
		     window);
  gtk_signal_connect(GTK_OBJECT(GTK_FILE_SELECTION(window)->cancel_button),
		     "clicked", 
		     GTK_SIGNAL_FUNC(destroyWidget),
		     window);
  gtk_widget_show(window);
}

/* end of insert.c */
